/*
    *
    * This file is a part of CorePad.
    * A document editor for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/


#include <QFile>
#include <QFileDialog>
#include <QShortcut>
#include <QFileInfo>
#include <QDate>
#include <QMessageBox>
#include <QDebug>
#include <QToolButton>
#include <QRegularExpression>
#include <QtConcurrent/QtConcurrent>

#include <cprime/infofunc.h>
#include <cprime/filefunc.h>
#include <cprime/themefunc.h>
#include <cprime/shareit.h>
#include <cprime/pinit.h>

#include "corepad.h"
#include "ui_corepad.h"
#include "ceditor.h"
#include "ccolorscheme.h"

// Get which color is bright
bool isBrightColor(QColor dst, QColor src)
{
	if (abs(dst.lightness() - src.lightness()) > 64) {
		return true;
	} else {
		return false;
	}
}

corepad::corepad(QWidget *parent) : QWidget(parent), ui(new Ui::corepad)
{
	ui->setupUi(this);

	loadSettings();
	startSetup();
	shotcuts();
}

corepad::~corepad()
{
	delete ui;
}

/**
 * @brief Loads application settings
 */
void corepad::loadSettings()
{
	sideViewSize = sm->value("CoreApps", "SideViewSize");
	sideViewIconSize = sm->value("CoreApps", "SideViewIconSize");
	toolBarIconSize = sm->value("CoreApps", "ToolBarIconSize");
	touch = sm->value("CoreApps", "TouchMode");
	textViewFont = sm->value("CorePad", "Font");
}

/**
 * @brief Setup ui elements
 */
void corepad::startSetup()
{
	// Used for numbering the document as untitled0.txt
	docCount = -1;

	// Set the color scheme to editor
	setColorScheme();

	// Set the transparent color to the main widget
	QPalette p(palette());
	p.setColor(QPalette::Base, Qt::transparent);
	this->setPalette(p);

	// Set stylesheet
	ui->sideView->setStyleSheet(CPrime::ThemeFunc::getSideViewStyleSheet());

	// all button with text
	for (QToolButton *b : ui->sideView->findChildren<QToolButton *>()) {
		b->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
		b->setFixedWidth(sideViewSize);
	}

	// Set sideview size
	ui->sideView->setFixedWidth(sideViewSize);

	ui->appTitle->setAttribute(Qt::WA_TransparentForMouseEvents);
	ui->appTitle->setFocusPolicy(Qt::NoFocus);
	this->resize(800, 500);

	// Set window size and extras
	if (touch) {
		this->setWindowState(Qt::WindowMaximized);

		if (not sm->value("CoreApps", "TabletMode")) {
			QToolButton *menu = new QToolButton;
			menu->setIcon(QIcon::fromTheme("application-menu"));
			menu->setIconSize(toolBarIconSize);

			ui->notes->setCornerWidget(menu, Qt::BottomLeftCorner);
			connect(menu, &QToolButton::clicked, this, &corepad::showSideView);
			connect(ui->appTitle, &QToolButton::clicked, this, &corepad::showSideView);

			ui->sideView->setVisible(0);
		}
	}

	// Startup page
	ui->page->setCurrentIndex(0);

	ui->topBar->setVisible(false);
	ui->cSave->setEnabled(false);
	ui->cSaveAs->setEnabled(false);
	ui->search->setEnabled(false);
	ui->addDate->setEnabled(false);
	ui->cPaste->setEnabled(false);
	ui->pinIt->setEnabled(false);
	ui->shareIt->setEnabled(false);

	connect(ui->cOpen, &QToolButton::clicked, this, &corepad::openFileDialog);
	connect(ui->addFiles, &QPushButton::clicked, this, &corepad::openFileDialog);
}

void corepad::shotcuts()
{
	QShortcut *shortcut;
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_N), this);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cNew_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_O), this);
	connect(shortcut, &QShortcut::activated, this, &corepad::openFileDialog);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_S), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cSave_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::SHIFT + Qt::Key_S), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cSaveAs_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Z), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cUndo_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Y), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cRedo_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_C), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cCopy_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_X), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cCut_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_V), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cPaste_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_Q), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::quitClicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_F), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_search_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::Key_Escape), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::hideSearch);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_W), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::closeCurrentTab);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::SHIFT + Qt::Key_W), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::closeAllTabs);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_T), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::on_cNew_clicked);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::ALT + Qt::Key_PageUp), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::previousTab);
	shortcut = new QShortcut(QKeySequence(Qt::ALT + Qt::Key_Left), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::previousTab);
	shortcut = new QShortcut(QKeySequence(Qt::CTRL + Qt::ALT + Qt::Key_PageDown), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::nextTab);
	shortcut = new QShortcut(QKeySequence(Qt::ALT + Qt::Key_Right), ui->app);
	connect(shortcut, &QShortcut::activated, this, &corepad::nextTab);
}

QIcon corepad::getTabIcon(QString filePath)
{
	QString defaultIconName = "document-new";

	QFile file(filePath);

	if (!file.exists()) {
		return QIcon::fromTheme(defaultIconName);
	}

	// Icon theme icons only works for breeze icons
	QMimeType mimeType = mimedb.mimeTypeForFile(filePath);
	QString iconName = mimeType.name().replace('/', '-');

	QIcon icon = QIcon::fromTheme(iconName);

	if (icon.isNull()) {
		return QIcon::fromTheme(defaultIconName);;
	}

	return icon;
}

void corepad::openFileDialog()
{
	QStringList fileP = QFileDialog::getOpenFileNames(this, tr("Open File"), QDir::homePath() + "/Documents", tr("All Files (*)"));

	if (!fileP.count()) {
		// Function from utilities.cpp
		CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Empty File Name", "Select a valid file.", this);
		return;
	} else {
		int failed = 0;

		for (QString path : fileP) {
			bool open = initializeNewTab(path);

			if (!open) {
				failed++;
			}
		}

		if (failed) {
			CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Can't open some file(s)", "Invalid file or corrupted.", this);
		} else {
			CPrime::InfoFunc::messageEngine("CorePad", "corepad", "File(s) Opened successfully", "Do your changes", this);
		}
	}
}

bool corepad::initializeNewTab(const QString &filePath)
{
	if (!filePath.count()) {
		CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Empty file path!!!", "Nothing to worry, select a valid file.", this);
		return false;
	}

	ui->page->setCurrentIndex(1);

	CEditor *edit = nullptr;

	// Check whether the file is already opened
	for (int i = 0; i < ui->notes->count(); i++) {
		CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

		if (tmp->workFilePath() == filePath) {
			edit = tmp;
			break;
		}
	}

	if (!edit) {
		QFont font = textViewFont;
		edit = new CEditor(filePath, scheme, this);
		edit->setFocusPolicy(Qt::ClickFocus);
		edit->setFont(font);
		edit->lineNumberArea()->setFont(font);

		connect(edit, &CEditor::copyAvailable, this, &corepad::updateShortcut);
		connect(edit->document(), &QTextDocument::undoAvailable, this, &corepad::updateShortcut);
		connect(edit->document(), &QTextDocument::redoAvailable, this, &corepad::updateShortcut);
		connect(edit, &CEditor::fileLoaded, this, &corepad::updateTab);
		connect(edit, &CEditor::unsavedChanges, this, &corepad::updateTab);
		ui->notes->addTab(edit, filePath);

		ui->notes->setTabIcon(ui->notes->count() - 1, getTabIcon(filePath));
		ui->notes->setIconSize(toolBarIconSize);
	}

	edit->loadFile(filePath);
	edit->setFocus();
	ui->notes->setCurrentWidget(edit);
	setWindowTitle(ui->notes->tabText(ui->notes->currentIndex()) + " - CorePad");
	return true;
}

void corepad::setColorScheme()
{
	scheme = new CColorScheme;

	QString theme = "";
	QColor mColor = scheme->getForegroundColor();
	QPalette pltt(palette());

	if (isBrightColor(pltt.color(QPalette::Window), mColor)) {
		if (pltt.color(QPalette::Window).lightness() < 128) {
			theme = "dark";
		} else {
			theme = "light";
		}
	} else {
		theme = "light";
	}

	qDebug() << "Theme Name " << theme;
	scheme = new CColorScheme(QString(":/syntax/%1.colorscheme").arg(theme));
}

void corepad::updateShortcut()
{
	CEditor *edit = currentEditor();

	// Check copy availability
	bool copy = edit->isCopyAvailable();
	ui->cCopy->setEnabled(copy);
	ui->cCut->setEnabled(copy);

	// Check undo availability
	ui->cUndo->setEnabled(edit->document()->isUndoAvailable());

	// Check redo availablity
	ui->cRedo->setEnabled(edit->document()->isRedoAvailable());
}

QStringList corepad::unsavedFiles()
{
	QStringList unsaved;

	for (int i = 0; i < ui->notes->count(); i++) {
		CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

		if (tmp->hasChange()) {
			unsaved << tmp->workFilePath();
		}
	}

	return unsaved;
}

void corepad::findS(QString searchS, bool reverse, QTextDocument::FindFlags flag)
{
	if (!searchS.isEmpty()) {
		CEditor *curr = currentEditor();

		bool found = curr->find(searchS, flag);

		if (!found) {
			curr->moveCursor(reverse ? QTextCursor::End : QTextCursor::Start);
			curr->find(searchS, flag);
		}
	}
}

void corepad::updateTab(QString file)
{
	CEditor *curr = nullptr;
	int currIndex = -1;

	for (int i = 0; i < ui->notes->count(); i++) {
		CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

		if (tmp->workFilePath() == file) {
			curr = tmp;
			currIndex = i;
			break;
		}
	}

	if (!curr) {
		return;
	}

	bool changes = curr->hasChange();
	ui->notes->setTabText(currIndex, (changes ? "*" : "") + QFileInfo(file).fileName());
	setWindowTitle(ui->notes->tabText(currIndex) + " - CorePad");
}

void corepad::newText()
{
	initializeNewTab(QString("new%1.txt").arg(++docCount));
}

void corepad::hideSearch()
{
	ui->topBar->setVisible(false);
	ui->searchHere->clear();

	CEditor *curr = currentEditor();

	if (!curr) {
		return;
	}

	QTextCursor tcur = curr->textCursor();
	tcur.clearSelection();
	curr->setTextCursor(tcur);
	curr->setFocus();
}

int corepad::navigateTab(bool next)
{
	int index = ui->notes->currentIndex();
	int len = ui->notes->count();

	if (next) {
		index++;

		if (index >= len) {
			index = 0;
		}
	} else {
		index--;

		if (index < 0) {
			index = len - 1;
		}
	}

	return index;
}

void corepad::nextTab()
{
	int index = navigateTab(true);
	ui->notes->setCurrentIndex(index);
	on_notes_currentChanged(index);
}

void corepad::previousTab()
{
	int index = navigateTab(false);
	ui->notes->setCurrentIndex(index);
	on_notes_currentChanged(index);
}

bool corepad::closeCurrentTab()
{
	return closeTab(ui->notes->currentIndex());
}

bool corepad::closeTab(int index)
{
	CEditor *edit = static_cast<CEditor *>(ui->notes->widget(index));

	if (edit == nullptr) {
		return false;
	}

	QString filePath = edit->workFilePath();

	if (edit->hasChange()) {
		int reply = QMessageBox::question(this, tr("Unsaved Changes?"),
										  QString(tr("This file has unsaved changes.\nDo you want to discard it?\n\n%1"))
										  .arg(filePath), QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

		if (reply != QMessageBox::Yes) {
			return false;
		}
	}

	if (QFileInfo(filePath).exists()) {
		qDebug() << filePath << "Saved to recent " << CPrime::InfoFunc::saveToRecent("corepad", filePath);
	}

	edit->deleteLater();
	ui->notes->removeTab(index);

	if (index == ui->notes->count()) {
		index -= 1;
	}

	on_notes_currentChanged(index);
	return true;
}

void corepad::closeAllTabs()
{
	for (int i = 0; i < ui->notes->count(); i++) {
		on_notes_tabCloseRequested(i);
	}
}

void corepad::quitClicked()
{
	int tabCount = ui->notes->count();
	int cc = 0;

	for (int i = 0; i < tabCount; i++) {
		if (closeTab(i - cc)) {
			cc++;
		}
	}

	if (ui->notes->count() == 0) {
		if (this->close()) {
			deleteLater();
		}
	} else {
		CPrime::InfoFunc::messageEngine("CorePad", "corepad", "Can't close", "There are still some changes needs to be saved.", this);
	}
}

void corepad::closeEvent(QCloseEvent *event)
{
	qDebug() << "Closing...";
	QStringList unsaved = unsavedFiles();

	if (unsaved.isEmpty()) {
		while (ui->notes->count()) {
			closeTab(ui->notes->count() - 1);
		}

		QWidget::closeEvent(event);
		return;
	}

	// Otherwise, ask the user what to do.
	QMessageBox::StandardButton but = QMessageBox::question(
										  this,
										  tr("Save Changes before closing?"),
										  QString(tr("There are unsaved changes.\nDo you want save them?\n\n%1")).arg(unsaved.join("\n")),
										  QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel,
										  QMessageBox::Yes);

	if (but == QMessageBox::Cancel) {
		event->ignore();
		return;
	} else if (but == QMessageBox::Yes) {
		// Save all opened files
		bool ok = true;

		for (int i = 0; i < ui->notes->count() && ok; i++) {
			CEditor *tmp = static_cast<CEditor *>(ui->notes->widget(i));

			if (tmp->hasChange()) {
				ok = ok && tmp->saveFile();
			}
		}

		if (!ok) {
			//cancelled by user
			event->ignore();
			return;
		}
	}

	QWidget::closeEvent(event);
}

void corepad::showSideView()
{
	if (touch) {
		if (ui->sideView->isVisible()) {
			ui->sideView->setVisible(0);
		} else {
			ui->sideView->setVisible(1);
		}
	}
}

void corepad::on_newTextW_clicked()
{
	newText();
}

void corepad::on_cNew_clicked()
{
	newText();
}

bool corepad::on_cSave_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return false;
	}

	bool save = curr->saveFile();

	// Update the tab icon while saved the file in a different file format
	ui->notes->setTabIcon(ui->notes->currentIndex(), getTabIcon(curr->workFilePath()));

	return save;
}

void corepad::on_cSaveAs_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	curr->saveFile(true);

	// Update the tab icon while saved the file in a different file format
	ui->notes->setTabIcon(ui->notes->currentIndex(), getTabIcon(curr->workFilePath()));
}

void corepad::on_cCopy_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	curr->copy();
}

void corepad::on_cPaste_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	curr->paste();
}

void corepad::on_cCut_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	curr->cut();
}

void corepad::on_cUndo_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	curr->undo();
}

void corepad::on_cRedo_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	curr->redo();
}

void corepad::on_addDate_clicked()
{
	currentEditor()->insertPlainText(QDate::currentDate().toString("dd/MM/yyyy"));
}

void corepad::on_search_clicked()
{
	CEditor *curr = currentEditor();

	if (curr == nullptr) {
		return;
	}

	if (ui->topBar->isVisible() && !curr->textCursor().selectedText().count()) {
		hideSearch();
	} else {
		ui->topBar->setVisible(true);
		ui->searchHere->setFocus();
		QString search = curr->textCursor().selectedText();

		if (search.count()) {
			ui->searchHere->setText(search);
		}
	}
}

void corepad::on_pinIt_clicked()
{
	QString workFilePath = currentEditor()->workFilePath();

	if (workFilePath.count()) {
		if (!QFileInfo(workFilePath).exists()) {
			CPrime::InfoFunc::messageEngine("CorePad", "corepad", "File not exists or saved", "Please choose a valid file or save the file first to pin it.", this);
			return;
		}

		pinit::callPinIt(this, workFilePath);
	}
}

void corepad::on_shareIt_clicked()
{
	QString workFilePath = currentEditor()->workFilePath();

	if (workFilePath.count()) {
		if (!QFileInfo(workFilePath).exists()) {
			// Function from utilities.cpp
			CPrime::InfoFunc::messageEngine("CorePad", "corepad", "File not exists or saved", "Please choose a valid file or save the file first to share it.", this);
			return;
		}

		shareit::shareFiles(this, QStringList() << workFilePath);
	}
}

void corepad::on_nextW_clicked()
{
	findS(ui->searchHere->text(), false);
}

void corepad::on_previousW_clicked()
{
	findS(ui->searchHere->text(), true, QTextDocument::FindBackward);
}

void corepad::on_replaceT_clicked()
{
	CEditor *curr = currentEditor();

	if (curr->textCursor().selectedText() == ui->searchHere->text()) {
		curr->insertPlainText(ui->replaceWith->text());
	}

	on_nextW_clicked();
}

void corepad::on_replaceAllT_clicked()
{
	CEditor *curr = currentEditor();
	QString selected = curr->textCursor().selectedText();

	if (selected.count()) {
		if (selected == ui->searchHere->text()) {
			curr->insertPlainText(ui->replaceWith->text());
		}
	}

	bool found = curr->find(ui->searchHere->text());

	while (found) {
		curr->insertPlainText(ui->replaceWith->text());
		found = curr->find(ui->searchHere->text());
	}
}

void corepad::on_searchHere_textChanged(const QString &arg1)
{
	if (!arg1.count()) {
		return;
	}

	CEditor *curr = currentEditor();

	if (curr->textCursor().atEnd()) {
		curr->moveCursor(QTextCursor::Start);
	} else {
		curr->moveCursor(QTextCursor::Start, QTextCursor::KeepAnchor);
	}

	QtConcurrent::run([ = ]() {
		int count = 0;
		QRegularExpression regex(arg1);
		QRegularExpressionMatch m = regex.match(curr->toPlainText());
		int index = m.capturedStart();

		while (index >= 0) {
			int l = m.capturedLength();
			count++;
			ui->searchCount->setText(QString("%1 matches").arg(count));
			m = regex.match(curr->toPlainText(), index + l);
			index = m.capturedStart();
		}

		qDebug() << "Count" << count;
	});

	findS(arg1, false);
}

void corepad::on_notes_currentChanged(int index)
{
	bool status;

	if (ui->notes->count() > 0) {
		status = true;

		updateShortcut();

		if (ui->topBar->isVisible()) {
			on_searchHere_textChanged(ui->searchHere->text());
		}

		setWindowTitle(ui->notes->tabText(index) + " - CorePad");
	} else {
		status = false;

		ui->topBar->setVisible(false);
		ui->cCopy->setEnabled(false);
		ui->cCut->setEnabled(false);
		ui->cUndo->setEnabled(false);
		ui->cRedo->setEnabled(false);

		setWindowTitle("CorePad");

		// Startup page
		ui->page->setCurrentIndex(0);
	}

	ui->cSave->setEnabled(status);
	ui->cSaveAs->setEnabled(status);
	ui->search->setEnabled(status);
	ui->addDate->setEnabled(status);
	ui->cPaste->setEnabled(status);
	ui->pinIt->setEnabled(status);
	ui->shareIt->setEnabled(status);
}

void corepad::on_notes_tabCloseRequested(int index)
{
	ui->notes->setCurrentIndex(index);
	closeTab(index);
}

CEditor *corepad::currentEditor()
{
	if (ui->notes->count() < 1) {
		return nullptr;
	}

	return static_cast<CEditor *>(ui->notes->currentWidget());
}

void corepad::sendFiles(const QStringList &paths)
{
	if (paths.count()) {
		for (QString str : paths) {
			if (CPrime::FileUtils::isFile(str)) {
				initializeNewTab(str);
			} else {
				qDebug() << "func(sendFiles) : Error : Invalid filepath got :" << str;
			}
		}
	}
}

void corepad::on_searchHere_returnPressed()
{
	on_nextW_clicked();
}
